__author__ = "elegans.io Ltd"
__email__ = "info@elegans.io"

import abc

from csrec.tools.observable import Observable
from csrec.tools.observable import observable
from csrec.exceptions import *

class DALBase(Observable):  # interface of the data abstraction layer
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        Observable.__init__(self)

    @abc.abstractmethod
    def init(self, **params):
        """
        initialization method

        exception: raise an InitializationException if any error occur

        :param params: dictionary of parameters
        """
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def get_init_parameters_description():
        """
        return a description of the supported options like the following:

        { "db_path", "the path of a database file previously initialized" }


        :return: a dictionary with the supported options
        """
        raise NotImplementedError

    @abc.abstractmethod
    @observable
    def insert_item(self, item_id, attributes):
        """
        insert a new item on datastore, if the item already exists replace it

        exception: raise an InsertException if any error occur

        :param item_id: item id
        :param attributes: a dictionary with item attributes e.g.
            {"author": "AA. VV.",
                "category":"horror",
                "subcategory":["splatter", "zombies"],
                ...
            }
        """
        raise NotImplementedError

    @abc.abstractmethod
    @observable
    def remove_item(self, item_id=None):
        """
        remove an item from datastore

        exception: raise a DeleteException if any error occur

        :param item_id: the item id to delete, if None remove all items
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_items(self, item_id=None):
        """
        get a dictionary of items

        exception: raise a GetException if any error occur

        :param item_id: the item id to get, if None get all items
        :return: a dictionary with one or more items:
            item_id0 : {"author": "AA. VV.",
                "category":"horror",
                "subcategory":["splatter", "zombies"],
                ...
            }
            ...
            item_idN : {"author": "AA. VV.",
                "category":"horror",
                "subcategory":["splatter", "zombies"],
                ...
            }

            an empty dictionary will be returned if the item was not found
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_items_iterator(self):
        """
        an iterator on items

        :return: an iterator on items
        """
        raise NotImplementedError

    @abc.abstractmethod
    def insert_social_action(self, user_id, user_id_to, code=3.0):
        """
        insert a new user id on datastore, for each user a list of actions will be mantained:
            user0: { 'user_0':3.0, ..., 'user_N':5.0}
            ...
            userN: { 'user_0':3.0, ..., 'user_N':5.0}

        exception: raise an InsertException if any error occur

        :param user_id: user id who make the action
        :param user_id_to: the user id destination of the action
        :param code: the code, default value is 3.0
        """
        raise NotImplementedError

    @abc.abstractmethod
    @observable
    def remove_social_action(self, user_id, user_id_to):
        """
        remove a social action from datastore

        exception: raise a DeleteException if any error occur

        :param user_id: user id who make the action
        :param user_id_to: the user id destination of the action
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_social_actions(self, user_id=None):
        """
        get the social actions

        exception: raise a GetException if any error occur

        :param user_id: user id, if None, return all social actions
        :return: a dictionary social actions performed BY users:
            user0: { 'user_1':3.0, ..., 'user_M':5.0}
            ...
            userN: { 'user_0':3.0, ..., 'user_M':5.0}
        """
        raise NotImplementedError

    @abc.abstractmethod
    @observable
    def insert_item_action(self, user_id, item_id, code=3.0, item_meaningful_info=None, only_info=False):
        """
        insert a new item code on datastore, for each user a list of ratings will be mantained:
            user0: { 'item_0':3.0, ..., 'item_N':5.0}
            ...
            userN: { 'item_0':3.0, ..., 'item_N':5.0}

        exception: raise an InsertException if any error occur

        :param user_id: user id
        :param item_id: item id
        :param code: the code, default value is 3.0
        """
        raise NotImplementedError

    @abc.abstractmethod
    @observable
    def remove_item_action(self, user_id, item_id):
        """
        remove a rating made by a user from the datastore

        exception: raise a DeleteException if any error occur

        :param user_id: user id
        :param item_id: item id
        :return: True if the operation was successfully executed or it does not exists, otherwise return False
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_item_actions(self, user_id=None):
        """
        get a dictionary with user's actions

        exception: raise a GetException if any error occur

        :param user_id: user id, if None returns actions for all users
        :return: a dictionary with all ratings:
            user0: { 'item_0':3.0, ..., 'item_N':5.0}
            ...
            userN: { 'item_0':3.0, ..., 'item_N':5.0}
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_item_actions_iterator(self):
        """
        get an iterator on item actions

        exception: raise a GetException if any error occur

        :return: an iterator on item ratings for each user:
            user0: { 'item_0':3.0, ..., 'item_N':5.0}
            ...
            userN: { 'item_0':3.0, ..., 'item_N':5.0}
        """
        raise NotImplementedError

    def get_item_ratings(self, item_id=None):
        """
        get ratings on items made by users

        exception: raise a GetException if any error occur

        :param item_id: an item id, if None returns the ratings for all items
        :return: a dictionary with ratings for each item
            item0: { 'user_0':3.0, ..., 'user_N':5.0}
            ...
            itemN: { 'user_0':3.0, ..., 'user_N':5.0}
        """
        raise NotImplementedError

    def get_info_used(self):
        """
        get the categories used

        exception: raise a GetException if any error occur

        :return: a set with the name of categories used
        """
        raise NotImplementedError

    def set_info_used(self, info_used):
        """
        insert a new category

        exception: raise an InsertException if any error occur

        :param info_used: the new category to use
        """
        raise NotImplementedError

    def remove_info_used(self, info_used=None):
        """
        remove a category from the datastore

        exception: raise a DeleteException if any error occur

        :param info_used: the category to be deleted, if None, reset all categories
        """
        raise NotImplementedError

    @abc.abstractmethod
    @observable
    def reconcile_user(self, old_user_id, new_user_id):
        """
        merge two users under the new user id, old user id will be removed
        for each item rated more than once, those rated by new_user_id will be kept

        exception: raise a MergeEntitiesException if any error occur

        :param old_user_id: old user id, raise an error if does not exists
        :param new_user_id: new user id, raise an error if does not exists
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_user_count(self):
        """
        get the number of users who rated items

        exception: raise a GetException if any error occur

        :return: the number of users
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_items_count(self):
        """
        get the number of items

        exception: raise a GetException if any error occur

        :return: the number of items
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_social_count(self):
        """
        Get the number of social actions
        :return: number of social actions
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_social_iterator(self):
        """
        iterator over social actions
        :return: iterator over social actions
        """
        raise NotImplementedError

    @abc.abstractmethod
    @observable
    def reset(self):
        """
        reset all data into the datastore

        exception: raise a DeleteException if any error occur
        """
        raise NotImplementedError

    @abc.abstractmethod
    @observable
    def serialize(self, filepath):
        """
        dump the datastore on file

        exception: raise a SerializeException if any error occur
        """
        raise NotImplementedError

    @abc.abstractmethod
    @observable
    def restore(self, filepath):
        """
        restore the datastore from file

        exception: raise a RestoreException if any error occur
        """
        raise NotImplementedError



__author__ = "elegans.io Ltd"
__email__ = "info@elegans.io"

# database abstraction layer factory

from csrec.exceptions import *


class Dal:
    @staticmethod
    def get_implemented_dal():
        """
        get the implemented dal
        :return: a set of implemented dal
        """

        implemented_dal = {
            'mem'
        }
        return implemented_dal

    @staticmethod
    def get_dal_supported_parameters(name):
        """
        get a list of supported parameters by a dal implementation and their description

        :param name: the name of any DAL implementation i.e. mem
        :return: the dictionary with the supported parameters and the description
        """

        if name == 'mem':  # in memory implementation of dal
            import csrec.mem_dal as mem_dal
            return mem_dal.Database.get_init_parameters_description()
        else:
            raise NotImplementedError

    @staticmethod
    def get_dal(name, **params):
        """
        database abstraction layer factory

        exception: raise a InitializationError if any error occur

        :param name: the name of any DAL implementation i.e. mem
        :param params: the dictionary of the parameters supported by the implementation. Refer to the
            help of the init process of the specific instance for the supported parameters.
        :return: an instance of the DAL implementation if the initialization was successfully executed
        """

        if name == 'mem':  # in memory implementation of dal
            import csrec.mem_dal as mem_dal
            try:
                dal_instance = mem_dal.Database()
            except InitializationException as exc:
                raise InitializationException("unable to initialize: " % name)
        else:
            raise NotImplementedError

        dal_instance.init(**params)

        return dal_instance



import pandas as pd
import numpy as np
from time import time
import logging
from csrec.tools.singleton import Singleton
from csrec import factory_dal

class Recommender(Singleton):
    """
    Cold Start Recommender
    """
    def __init__(self, dal_name='mem', dal_params={}, max_rating=5, log_level=logging.INFO):
        # Logger initialization
        self.logger = logging.getLogger("csrc")
        self.logger.setLevel(log_level)
        ch = logging.StreamHandler()
        ch.setLevel(log_level)
        self.logger.addHandler(ch)
        self.logger.debug("============ Logger initialized ================")

        # initialization of datastore attribute
        self.db = factory_dal.Dal.get_dal(name=dal_name, **dal_params)  # instantiate an in memory database

        # registering callback functions for datastore events
        self.db.register(self.db.serialize, self.on_serialize)
        self.db.register(self.db.restore, self.on_restore)

        # Algorithm's specific attributes
        self._items_cooccurrence = pd.DataFrame  # cooccurrence of items
        self.cooccurrence_updated = 0.0
        # Info in item_meaningful_info with whom some user has actually interacted
        self._categories_cooccurrence = {}  # cooccurrence of categories

        # categories --same as above, but separated as they are not always available
        self.items_by_popularity = []  # can be recomputed on_restore
        self.last_serialization_time = 0.0  # Time of data backup
        # configurations:
        self.max_rating = max_rating

    def on_serialize(self, filepath, return_value):
        if return_value is None or return_value:
            self.last_serialization_time = time()
        else:
            self.logger.error("[on_serialize] data backup failed on file %s, last successful backup at: %f" %
                              (filepath,
                               self.last_serialization_time))

    def on_restore(self, filepath, return_value):
        if return_value is not None and not return_value:
            self.logger.error("[on_restore] restore from serialized data fail: ", filepath)
        else:
            self._create_cooccurrence()

    def _create_cooccurrence(self):
        """
        Create or update the co-occurrence matrix
        :return:
        """
        all_ratings = self.db.get_item_actions()
        df = pd.DataFrame(all_ratings).fillna(0).astype(int)  # convert dictionary to pandas dataframe

        # calculate co-occurrence matrix
        # sometime will print the warning: "RuntimeWarning: invalid value encountered in true_divide"
        # use np.seterr(divide='ignore', invalid='ignore') to suppress this warning
        df_items = (df / df).replace(np.inf, 0).replace(np.nan, 0)  # calculate co-occurrence matrix and normalize to 1
        co_occurrence = df_items.fillna(0).dot(df_items.T)
        self._items_cooccurrence = co_occurrence

        # update co-occurrence matrix for items categories
        df_n_cat_item = {}

        info_used = self.db.get_info_used()
        if len(info_used) > 0:
            n_categories_item_ratings = self.db.get_n_categories_item_ratings()
            for i in info_used:
                df_n_cat_item[i] = pd.DataFrame(n_categories_item_ratings.get(i)).fillna(0).astype(int)

            for i in info_used:
                if type(df_n_cat_item.get(i)) == pd.DataFrame:
                    df_n_cat_item[i] = (df_n_cat_item[i] / df_n_cat_item[i]).replace(np.nan, 0)
                    self._categories_cooccurrence[i] = df_n_cat_item[i].T.dot(df_n_cat_item[i])

        self.cooccurrence_updated = time()

    def compute_items_by_popularity(self):
        """
        As per name, get self.items_by_popularity
        :return: None
        """
        df_item = pd.DataFrame(self.db.get_item_actions()).T.fillna(0).astype(int).sum()
        df_item.sort_values(inplace=True, ascending=False)
        pop_items = list(df_item.index)
        all_items = set(self.db.get_items().keys())
        self.items_by_popularity = (pop_items + list(all_items - set(pop_items)))

    def get_recommendations(self, user_id, max_recs=50, fast=False, algorithm='item_based'):
        """
        algorithm item_based:
            - Compute recommendation to user using item co-occurrence matrix (if the user
            rated any item...)
            - If there are less than max_recs recommendations, the remaining
            items are given according to popularity. Scores for the popular ones
            are given as score[last recommended]*index[last recommended]/n
            where n is the position in the list.
            - Recommended items above receive a further score according to categories
        :param user_id: the user id as in the collection of 'users'
        :param max_recs: number of recommended items to be returned
        :param fast: Compute the co-occurrence matrix only if it is one hour old or
                     if matrix and user vector have different dimension
        :return: list of recommended items
        """
        user_id = str(user_id).replace('.', '')
        df_tot_cat_user = {}
        df_n_cat_user = {}
        rec = pd.Series()
        user_has_rated_items = False  # has user rated some items?
        rated_infos = []  # user has rated the category (e.g. the category "author" etc)
        df_user = None
        if self.db.get_item_actions(user_id=user_id):  # compute item-based rec only if user has rated smt
            user_has_rated_items = True
            # Just take user_id for the user vector
            df_user = pd.DataFrame(self.db.get_item_actions()).fillna(0).astype(int)[[user_id]]
        info_used = self.db.get_info_used()
        if len(info_used) > 0:
            tot_categories_user_ratings = self.db.get_tot_categories_user_ratings()
            n_categories_user_ratings = self.db.get_n_categories_user_ratings()
            for i in info_used:
                if i in tot_categories_user_ratings and user_id in tot_categories_user_ratings[i]:
                    rated_infos.append(i)
                    df_tot_cat_user[i] =\
                        pd.DataFrame(tot_categories_user_ratings.get(i)).fillna(0).astype(int)[[user_id]]
                    df_n_cat_user[i] =\
                        pd.DataFrame(n_categories_user_ratings.get(i)).fillna(0).astype(int)[[user_id]]

        if user_has_rated_items:
            if not fast or (time() - self.cooccurrence_updated > 1800):
                self._create_cooccurrence()
            try:
                # this might fail if the cooccurrence was not updated (fast)
                # and the user rated a new item.
                # In this case the matrix and the user-vector have different
                # dimension
                # print("DEBUG [get_recommendations] Trying cooccurrence dot df_user")
                # print("DEBUG [get_recommendations] _items_cooccurrence: %s", self._items_cooccurrence)
                # print("DEBUG [get_recommendations] df_user: %s", df_user)
                rec = self._items_cooccurrence.T.dot(df_user[user_id])
                # self.logger.debug("[get_recommendations] Rec worked: %s", rec)
            except:
                self.logger.debug("[get_recommendations] 1st rec production failed, calling _create_cooccurrence.")
                self._create_cooccurrence()
                rec = self._items_cooccurrence.T.dot(df_user[user_id])
                self.logger.debug("[get_recommendations] Rec: %s", rec)
            # Sort by cooccurrence * rating:
            rec.sort_values(inplace=True, ascending=False)

            # If necessary, add popular items
            if len(rec) < max_recs:
                if not fast or len(self.items_by_popularity) == 0:
                    self.compute_items_by_popularity()
                for v in self.items_by_popularity:
                    if len(rec) == max_recs:
                        break
                    elif v not in rec.index:
                        n = len(rec)
                        # supposing score goes down according to Zipf distribution
                        rec.set_value(v, rec.values[n - 1]*n/(n+1.))

        else:
            if not fast or len(self.items_by_popularity) == 0:
                self.compute_items_by_popularity()
            for i, v in enumerate(self.items_by_popularity):
                if len(rec) == max_recs:
                    break
                rec.set_value(v, self.max_rating / (i+1.))  # As comment above, starting from max_rating
#        print("DEBUG [get_recommendations] Rec after item_based or not: %s", rec)

        # Now, the worse case we have is the user has not rated, then rec=popular with score starting from max_rating
        # and going down as 1/i

        # User info on rated categories (in info_used)
        global_rec = rec.copy()
        if len(info_used) > 0:
            cat_rec = {}
            if not fast or (time() - self.cooccurrence_updated > 1800):
                self._create_cooccurrence()
            for cat in rated_infos:
                # get average rating on categories
                user_vec = df_tot_cat_user[cat][user_id] / df_n_cat_user[cat][user_id].replace(0, 1)
                # print("DEBUG [get_recommendations]. user_vec:\n", user_vec)
                try:
                    cat_rec[cat] = self._categories_cooccurrence[cat].T.dot(user_vec)
                    cat_rec[cat].sort_values(inplace=True, ascending=False)
                    #print("DEBUG [get_recommendations] cat_rec (try):\n %s", cat_rec)
                except:
                    self._create_cooccurrence()
                    cat_rec[cat] = self._categories_cooccurrence[cat].T.dot(user_vec)
                    cat_rec[cat].sort(ascending=False)
                    #print("DEBUG [get_recommendations] cat_rec (except):\n %s", cat_rec)
                for item_id, score in rec.iteritems():
                    #print("DEBUG [get_recommendations] rec_item_id: %s", k)
                    try:
                        item = self.db.get_items(item_id=item_id)
                        item_info_value = item.get(cat, "")

                        #print("DEBUG get_recommendations. item value for %s: %s", cat, item_info_value)
                        # In case the info value is not in cat_rec (as it can obviously happen
                        # because a rec'd item coming from most popular can have the value of
                        # an info (author etc) which is not in the rec'd info
                        if item_info_value:
                            global_rec[item_id] = score + cat_rec.get(cat, {}).get(item_info_value, 0)
                    except Exception as e:
                        self.logger.error("item %s, category %s", item_id, cat)
                        logging.exception(e)
        global_rec.sort_values(inplace=True, ascending=False)
#        print("DEBUG [get_recommendations] global_rec:\n %s", global_rec)

        if user_has_rated_items:
            # If the user has rated all items, return an empty list
            rated = df_user[user_id] != 0
            # rated.get is correct (pycharm complains, knows no pandas)
            items = [i for i in global_rec.index if not rated.get(i, False)]
            if items:
                return items[:max_recs]
            else:
                return []
        else:
            try:
                return list(global_rec.index)[:max_recs]
            except:
                return None



__author__ = "Angelo Leto"
__email__ = "angleto@gmail.com"

import math

def ShannonEntropy(pArray):
    """
    calculate shannon entropy for an array of elements

    :param pArray: array of elements
    :return: the shannon entropy value
    """
    N = float(sum(pArray))
    H = 0
    for i in range(0,len(pArray)):
        try:
            H += ( (float(pArray[i])/N) * math.log((float(pArray[i])/N) + (1 if pArray[i] == 0 else 0)) )
        except:
            H += 0
    return H

def LogLikelihoodRatio(pKTable):
    """
    :param pKTable: table of 2x2 elements stored in a 4 elements vector, with conditional probability
    :return: the log likelihood ratio, see http://tdunning.blogspot.it/2008/03/surprise-and-coincidence.html
    """
    try:
        v = 2 * sum(pKTable) * (ShannonEntropy(pKTable) -
                                ShannonEntropy([pKTable[0] + pKTable[1], pKTable[2] + pKTable[3]]) -
                                ShannonEntropy([pKTable[0] + pKTable[2], pKTable[1] + pKTable[3]]))
    except:
        v = 0
    return v





__author__ = "elegans.io Ltd"
__email__ = "info@elegans.io"


__base_error_code__ = 110

from collections import defaultdict

import pickle  # serialization library
from csrec.dal import DALBase
from csrec.tools.singleton import Singleton
from csrec.tools.observable import observable
import json

from csrec.exceptions import *


class Database(DALBase, Singleton):
    def __init__(self):
        DALBase.__init__(self)

        self.__params_dictionary = {}  # abstraction layer initialization parameters

        self.items_tbl = {}  # table with items

        self.users_ratings_tbl = {}  # table with users rating
        self.items_ratings_tbl = {}  # table with items rating

        self.users_social_tbl = {}  # table with action user-user
        self.info_used = set()

        self.tot_categories_user_ratings = {}  # sum of all ratings
        self.tot_categories_item_ratings = {}  # ditto
        self.n_categories_user_ratings = {}  # number of ratings
        self.n_categories_item_ratings = {}

    def init(self, **params):
        if not params:
            params = {}
        try:
            self.__params_dictionary.update(params)
        except Exception as e:
            e_message = "error during initialization"
            raise InitializationException(e_message + " : " + e.message)

    @staticmethod
    def get_init_parameters_description():
        param_description = {}
        return param_description

    @observable
    def insert_item(self, item_id, attributes=None):
        """
        insert a new item on datastore, if the item already exists replace it

        exception: raise an InsertException if any error occur

        :param item_id: item id
        :param attributes: a dictionary with item attributes e.g.
            {"author": "AA. VV.",
                "category":"horror",
                "subcategory":["splatter", "zombies"],
                ...
            }
        """
        if attributes is not None:
            for k, v in attributes.iteritems():
                try:
                    v = json.loads(v)
                except ValueError:
                    pass

                if not isinstance(v, list):
                    values = [v]
                else:
                    values = v

                self.items_tbl.setdefault(item_id, {})[k] = values
        else:
            self.items_tbl[item_id] = {}
        return True

    @observable
    def remove_item(self, item_id=None):
        """
        remove an item from datastore

        exception: raise a DeleteException if any error occur

        :param item_id: the item id to delete, if None remove all items
        """
        if item_id is not None:
            try:
                del self.items_tbl[item_id]
            except KeyError:
                pass
        else:
            self.items_tbl.clear()

    def get_items(self, item_id=None):
        """
        get a dictionary of items

        exception: raise a GetException if any error occur

        :param item_id: the item id to get, if None get all items
        :return: a dictionary with one or more items:
            item_id0 : {"author": "AA. VV.",
                "category":"horror",
                "subcategory":["splatter", "zombies"],
                ...
            }
            ...
            item_idN : {"author": "AA. VV.",
                "category":"horror",
                "subcategory":["splatter", "zombies"],
                ...
            }

            an empty dictionary will be returned if the item was not found
        """

        if item_id is not None:
            return {item_id: self.items_tbl.get(item_id)}
        else:
            return self.items_tbl

    def get_items_iterator(self):
        """
        an iterator on items

        :return: an iterator on items
        """
        return self.items_tbl.iteritems()

    @observable
    def insert_social_action(self, user_id, user_id_to, code=3.0):
        """
        insert a new user id on datastore, for each user a list of actions will be maintained:
            user0: { 'user_0':3.0, ..., 'user_N':5.0}
            ...
            userN: { 'user_0':3.0, ..., 'user_N':5.0}

        exception: raise an InsertException if any error occur

        :param user_id: user id who make the action
        :param user_id_to: the user id destination of the action
        :param code: the code, default value is 3.0
        """
        self.users_social_tbl.setdefault(user_id, {})[user_id_to] = code

    @observable
    def remove_social_action(self, user_id, user_id_to):
        """
        remove a social action from datastore

        exception: raise a DeleteException if any error occur

        :param user_id: user id who make the action
        :param user_id_to: the user id destination of the action
        """
        try:
            del self.users_social_tbl[user_id][user_id_to]
        except KeyError:
            pass

    def get_social_actions(self, user_id=None):
        """
        get the social actions

        exception: raise a GetException if any error occur

        :param user_id: user id, if None, return all social actions
        :return: a dictionary social actions performed BY users:
            user0: { 'user_1':3.0, ..., 'user_M':5.0}
            ...
            userN: { 'user_0':3.0, ..., 'user_M':5.0}
        """
        if user_id is not None:
            social_actions = self.users_social_tbl.get(user_id)
            if social_actions is None:
                return {}
            else:
                return {user_id: social_actions}
        else:
            return self.users_social_tbl

    @observable
    def insert_item_action(self, user_id, item_id, code=3.0, item_meaningful_info=None, only_info=False):
        """
        insert a new item code on datastore, for each user a list of ratings will be maintained:
            user0: { 'item_0':3.0, ..., 'item_N':5.0}
            ...
            userN: { 'item_0':3.0, ..., 'item_N':5.0}

        exception: raise an InsertException if any error occur

        :param user_id: user id
        :param item_id: item id
        :param code: the code, default value is 3.0
        :param item_meaningful_info: list of info to be considered, e.g. ['Author', 'tags']
        :param only_info: should only the info, and not the item, be considered
        """
        if item_meaningful_info is None:
            item_meaningful_info = []

        # If only_info==True, only the self.item_meaningful_info's are put in the co-occurrence, not item_id.
        # This is necessary when we have for instance a "segmentation page" where we propose
        # well known items to get to know the user. If s/he select "Harry Potter" we only want
        # to retrieve the info that s/he likes JK Rowling, narrative, magic etc

        # Now fill the dicts or the db collections if available
        user_id = str(user_id).replace('.', '')

        item = self.get_items(item_id=item_id)[item_id]
        if item is not None:
            for info in item_meaningful_info:
                values = item.get(info)
                if values is not None:
                    self.set_info_used(info)

                    # we cannot set the rating, because we want to keep the info
                    # that a user has read N books of, say, the same author,
                    # category etc.
                    # We could, but won't, sum all the ratings and count the a result as "big rating".
                    # We won't because reading N books of author A and rating them 5 would be the same
                    # as reading 5*N books of author B and rating them 1.
                    # Therefore we take the average because --
                    # 1) we don't want ratings for category to skyrocket
                    # 2) if a user changes their idea on rating a book, it should not add up.
                    # Average is not perfect, but close enough.
                    #
                    # Take total number of ratings and total rating:
                    for value in values:
                        self.tot_categories_user_ratings.setdefault(info, {}).setdefault(user_id, {}).setdefault(value, 0)
                        self.tot_categories_user_ratings[info][user_id][value] += int(code)
                        self.n_categories_user_ratings.setdefault(info, {}).setdefault(user_id, {}).setdefault(value, 0)
                        self.n_categories_user_ratings[info][user_id][value] += 1
                        # same for items
                        self.tot_categories_item_ratings.setdefault(info, {}).setdefault(value, {}).setdefault(user_id, 0)
                        self.tot_categories_item_ratings[info][value][user_id] += int(code)
                        self.n_categories_item_ratings.setdefault(info, {}).setdefault(value, {}).setdefault(user_id, 0)
                        self.n_categories_item_ratings [info][value][user_id] += 1

        else:
            self.insert_item(item_id=item_id)
        if not only_info:
                self.users_ratings_tbl.setdefault(user_id, {})[item_id] = code
                self.items_ratings_tbl.setdefault(item_id, {})[user_id] = code

    @observable
    def remove_item_action(self, user_id, item_id):
        """
        remove a rating made by a user from the datastore

        exception: raise a DeleteException if any error occur

        :param user_id: user id
        :param item_id: item id
        :return: True if the operation was successfully executed or it does not exists, otherwise return False
        """
        try:
            del self.users_ratings_tbl[user_id][item_id]
        except KeyError:
            pass

        try:
            del self.items_ratings_tbl[item_id][user_id]
        except KeyError:
            pass

    def get_item_actions(self, user_id=None):
        """
        get a dictionary with user's actions

        exception: raise a GetException if any error occur

        :param user_id: user id, if None returns actions for all users
        :return: a dictionary with all ratings:
            user0: { 'item_0':3.0, ..., 'item_N':5.0}
            ...
            userN: { 'item_0':3.0, ..., 'item_N':5.0}
        """
        if user_id is not None:
            item_actions = self.users_ratings_tbl.get(user_id)
            if item_actions is None:
                return {}
            else:
                return {user_id: item_actions}
        else:
            return self.users_ratings_tbl

    def get_item_actions_iterator(self):
        """
        get an iterator on item actions

        exception: raise a GetException if any error occur

        :return: an iterator on item ratings for each user:
            user0: { 'item_0':3.0, ..., 'item_N':5.0}
            ...
            userN: { 'item_0':3.0, ..., 'item_N':5.0}
        """
        return self.users_ratings_tbl.iteritems()

    def get_item_ratings(self, item_id=None):
        """
        get ratings on items made by users

        exception: raise a GetException if any error occur

        :param item_id: an item id, if None returns the ratings for all items
        :return: a dictionary with ratings for each item
            item0: { 'user_0':3.0, ..., 'user_N':5.0}
            ...
            itemN: { 'user_0':3.0, ..., 'user_N':5.0}
        """
        if item_id is not None:
            users_actions = self.items_ratings_tbl.get(item_id)
            if users_actions is None:
                return {}
            else:
                return {item_id: users_actions}
        else:
            return self.items_ratings_tbl

    def get_info_used(self):
        """
        get the categories used

        exception: raise a GetException if any error occur

        :return: a set with the name of categories used
        """
        return self.info_used

    def set_info_used(self, info_used):
        """
        insert a new category

        exception: raise an InsertException if any error occur

        :param info_used: the new category to use
        """
        self.info_used.add(info_used)

    def remove_info_used(self, info_used=None):
        """
        remove a category from the datastore

        exception: raise a DeleteException if any error occur

        :param info_used: the category to be deleted, if None, reset all categories
        """
        if info_used:
            try:
                self.info_used.remove(info_used)
            except KeyError:
                pass
        else:
            self.info_used.clear()

    @observable
    def reconcile_user(self, old_user_id, new_user_id):
        """
        merge two users under the new user id, old user id will be removed
        for each item rated more than once, those rated by new_user_id will be kept

        exception: raise a MergeEntitiesException if any error occur

        :param old_user_id: old user id, raise an error if does not exists
        :param new_user_id: new user id, raise an error if does not exists
        """
        #  verifying that both users exists
        if old_user_id not in self.users_ratings_tbl:
            e_message = "unable to reconcile old user id does not exists: %s" % str(old_user_id)
            raise MergeEntitiesException(e_message)

        if new_user_id not in self.users_ratings_tbl:
            e_message = "unable to reconcile new user id does not exists: %s" % str(new_user_id)
            raise MergeEntitiesException(e_message)

        if old_user_id == new_user_id:
            e_message = "users to be reconcile are the same: %s" % str(new_user_id)
            raise MergeEntitiesException(e_message)

        # updating ratings
        old_user_actions = self.users_ratings_tbl[old_user_id]
        self.users_ratings_tbl[new_user_id].update(old_user_actions)
        del self.users_ratings_tbl[old_user_id]

        # replacing all ratings of the user
        for i, r in old_user_actions.items():
            try:
                del self.items_ratings_tbl.setdefault(i, {})[old_user_id]
            except KeyError:
                pass
            self.items_ratings_tbl.setdefault(i, {})[new_user_id] = r

        # updating the social stuff
        old_social_dict = {}
        try:
            old_social_dict = self.users_social_tbl[old_user_id]
            del self.users_social_tbl[old_user_id]
        except KeyError:
            pass

        new_social_dict = {}
        try:
            new_social_dict = self.users_social_tbl[new_user_id]
        except KeyError:
            pass

        old_social_dict.update(new_social_dict)
        self.users_social_tbl[new_user_id] = old_social_dict

        for category in self.info_used:
            if old_user_id in self.tot_categories_user_ratings[category]:
                tot_curr_cat_values = self.tot_categories_user_ratings[category][old_user_id]
                del self.tot_categories_user_ratings[category][old_user_id]
                if new_user_id not in self.tot_categories_user_ratings[category]:
                    self.tot_categories_user_ratings[category][new_user_id] = tot_curr_cat_values
                else:
                    for value, tot_code in tot_curr_cat_values.iteritems():
                        self.tot_categories_user_ratings[category][new_user_id].setdefault(value, 0)
                        self.tot_categories_user_ratings[category][new_user_id][value] += tot_code

                n_curr_cat_values = self.n_categories_user_ratings[category][old_user_id]
                del self.n_categories_user_ratings[category][old_user_id]
                if new_user_id not in self.n_categories_user_ratings[category]:
                    self.n_categories_user_ratings[category][new_user_id] = n_curr_cat_values
                else:
                    for value, n_code in n_curr_cat_values.iteritems():
                        self.n_categories_user_ratings[category][new_user_id].setdefault(value, 0)
                        self.n_categories_user_ratings[category][new_user_id][value] += n_code

            for v in self.tot_categories_item_ratings[category]:
                if old_user_id in self.tot_categories_item_ratings[category][v]:
                    tot_curr_cat_item_values = self.tot_categories_item_ratings[category][v][old_user_id]
                    del self.tot_categories_item_ratings[category][v][old_user_id]
                    self.tot_categories_item_ratings[category][v].setdefault(new_user_id, 0)
                    self.tot_categories_item_ratings[category][v][new_user_id] += tot_curr_cat_item_values

                    tot_curr_cat_item_values = self.n_categories_item_ratings[category][v][old_user_id]
                    del self.n_categories_item_ratings[category][v][old_user_id]
                    self.n_categories_item_ratings[category][v].setdefault(new_user_id, 0)
                    self.n_categories_item_ratings[category][v][new_user_id] += 1


    def get_user_count(self):
        """
        get the number of users who rated items

        exception: raise a GetException if any error occur

        :return: the number of users
        """
        return len(self.users_ratings_tbl)

    def get_items_count(self):
        """
        get the number of items

        exception: raise a GetException if any error occur

        :return: the number of items
        """
        return len(self.items_tbl)

    def get_social_count(self):
        """
        Get the number of social actions
        :return: number of social actions
        """
        tot_social = 0
        try:
            for user in self.users_social_tbl:
                tot_social += len(self.users_social_tbl.get(user, {}).keys())
        except:
            tot_social = 0
        return tot_social

    def get_social_iterator(self):
        for user in self.users_social_tbl:
            yield {user: self.users_social_tbl[user]}

    @observable
    def reset(self):
        """
        reset all data into the datastore

        exception: raise a DeleteException if any error occur
        """
        self.items_tbl.clear()
        self.users_ratings_tbl.clear()
        self.items_ratings_tbl.clear()
        self.users_social_tbl.clear()
        self.tot_categories_user_ratings.clear()
        self.tot_categories_item_ratings.clear()
        self.n_categories_user_ratings.clear()
        self.info_used.clear()

    @observable
    def serialize(self, filepath):
        """
        dump the datastore on file

        exception: raise a SerializeException if any error occur
        """
        # Write chunks of text data
        try:
            with open(filepath, 'wb') as f:
                data_to_serialize = {'items': self.items_tbl,
                                     'users_ratings': self.users_ratings_tbl,
                                     'items_ratings': self.items_ratings_tbl,
                                     'user_social': self.users_social_tbl,
                                     'tot_categories_user_ratings': self.tot_categories_user_ratings,
                                     'tot_categories_item_ratings': self.tot_categories_item_ratings,
                                     'n_categories_user_ratings': self.n_categories_user_ratings,
                                     'info_used': self.info_used
                                     }
                pickle.dump(data_to_serialize, f)
        except Exception as e:
            e_message = "unable to serialize data to file: %d" % (__base_error_code__ + 1)
            raise SerializeException(e_message + " : " + e.message)

    @observable
    def restore(self, filepath):
        """
        restore the datastore from file

        exception: raise a RestoreException if any error occur
        """
        # Write chunks of text data
        try:
            with open(filepath, 'rb') as f:
                data_from_file = pickle.load(f)
                self.items_tbl = data_from_file['items']
                self.users_ratings_tbl = data_from_file['users_ratings']
                self.items_ratings_tbl = data_from_file['items_ratings']
                self.users_social_tbl = data_from_file['user_social']
                self.tot_categories_user_ratings = data_from_file['tot_categories_user_ratings']
                self.tot_categories_item_ratings = data_from_file['tot_categories_item_ratings']
                self.n_categories_user_ratings = data_from_file['n_categories_user_ratings']
                self.info_used = data_from_file['info_used']
        except Exception as e:
            e_message = "unable to load data from file: %d" % (__base_error_code__ + 2)
            raise RestoreException(e_message + " : " + e.message)

    def get_tot_categories_user_ratings(self):
        return self.tot_categories_user_ratings

    def get_tot_categories_item_ratings(self):
        return self.tot_categories_item_ratings

    def get_n_categories_item_ratings(self):
        return self.n_categories_item_ratings

    def get_n_categories_user_ratings(self):
        return self.n_categories_user_ratings



__author__ = "elegans.io Ltd"
__email__ = "info@elegans.io"


class DatabaseOperationException(Exception):
    """
    Exception used to report generic database error
    """
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class IntegrityViolationException(Exception):
    """
    Exception used for integrity violations, e.g. item already in database
    """
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class GetException(Exception):
    """
    Exception used to report problems getting data from db
    """
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class InsertException(Exception):
    """
    Exception used to report problems inserting data on db
    """
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class SerializeException(Exception):
    """
    Exception used to report problems during data serialization
    """
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class RestoreException(Exception):
    """
    Exception used to report problems restoring data on db from a serialized archive
    """
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class MergeEntitiesException(Exception):
    """
    Exception used to report problems merging data on db, e.g. user reconciliation
    """
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class DeleteException(Exception):
    """
    Exception used to report problems removing data from db
    """
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class InitializationException(Exception):
    """
    Exception used to report problems during DAL initialization initialization
    """
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class BadParametersException(Exception):
    """
    Exception used to report that initialization has failed because of unsupported or wrong parameters
    """
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)



__version__ = '0.4.1'
__author__ = 'elegans.io Ltd'
__license__ = 'GPL v2'
__maintainer__ = 'elegans-io'
__email__ = 'info@elegans.io'

from csrec.factory_dal import Dal
from csrec.recommender import Recommender



__author__ = "elegans.io Ltd"
__email__ = "info@elegans.io"

import abc
from functools import wraps


def observable(function):
    """
    observable decorator
    :param function:
    :return:
    """
    @wraps(function)
    def newf(*args, **kwargs):
        return_value = function(*args, **kwargs)
        called_class = args[0]

        try:
            observers = called_class.observers[function]
        except KeyError:
            pass
        else:
            for o in observers:
                kwargs['return_value'] = return_value
                o(**kwargs)
        return return_value
    return newf


class Observable(object):
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        self.observers = {}

    def register(self, function, observer):
        """
        register an observer to a specific event

        :param function: the function to observe
        :param observer: the observer function
        :return:
        """
        event = function.__name__
        if event not in self.observers:
            self.observers[event] = {}

        try:
            self.observers[event][observer] = 1
        except KeyError:
            op = False
        else:
            op = True
        return op

    def unregister(self, function, observer):
        event = function.__name__
        try:
            del self.observers[event][observer]
        except KeyError:
            return False
        else:
            return True

    def unregister_all(self):
        self.observers = {}


from setuptools import setup, find_packages
from codecs import open  # To use a consistent encoding
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
long_description = open('README.md').read()


def setup_package():

    build_requires = ['numpy>=1.10.1', 'pandas>=0.17.1']

    metadata = dict(
        name='cold-start-recommender',

        # Versions should comply with PEP440.  For a discussion on single-sourcing
        # the version across setup.py and the project code, see
        # https://packaging.python.org/en/latest/single_source_version.html
        version='0.4.1',
        description='In-memory recommender for recommendations produced on-the-fly',
        long_description=long_description,

        # The project's main homepage.
        url='https://github.com/elegans-io/cold-start-recommender',

        # Author details
        author='elegans.io Ltd',
        author_email='info@elegans.io',

        # Choose your license
        license='GPL v2',

        # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
        classifiers=[
            # How mature is this project? Common values are
            #   3 - Alpha
            #   4 - Beta
            #   5 - Production/Stable
            'Development Status :: 4 - Beta',

            # Indicate who your project is intended for
            'Intended Audience :: Information Technology',
            'Topic :: Scientific/Engineering :: Artificial Intelligence',

            # Pick your license as you wish (should match "license" above)
            'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',

            # Specify the Python versions you support here. In particular, ensure
            # that you indicate whether you support Python 2, Python 3 or both.
            'Programming Language :: Python :: 2.7',
        ],

        # What does your project relate to?
        keywords='recommendations, recommender,recommendation engine',

        # You can just specify the packages manually here if your project is
        # simple. Or you can use find_packages().
        #packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
        packages=['csrec'],

        # List run-time dependencies here.  These will be installed by pip when your
        # project is installed. For an analysis of "install_requires" vs pip's
        # requirements files see:
        # https://packaging.python.org/en/latest/requirements.html
        # See also https://github.com/scipy/scipy/blob/master/setup.py (malemi)
        setup_requires=build_requires,
        install_requires=build_requires,

        # List additional groups of dependencies here (e.g. development dependencies).
        # You can install these using the following syntax, for example:
        # $ pip install -e .[dev,test]
        # extras_require={
        #     'dev': ['check-manifest'],
        #     'test': ['coverage'],
        # },

        # If there are data files included in your packages that need to be
        # installed, specify them here.  If using Python 2.6 or less, then these
        # have to be included in MANIFEST.in as well.
        package_data={
            'csrec': ['*.cl', '*.py']
        },

        include_package_data=True

        # To provide executable scripts, use entry points in preference to the
        # "scripts" keyword. Entry points provide cross-platform support and allow
        # pip to create the appropriate form of executable for the target platform.
        # entry_points={
        #     'console_scripts': [
        #         'sample=sample:main',
        #     ],
        # },
    )

    setup(**metadata)

if __name__ == '__main__':
    setup_package()
